/*
nsisFile -- NSIS plugin for file content manipulation
Web site: http://wiz0u.free.fr/prog/nsisFile

Copyright (c) 2010 Olivier Marcoux

This software is provided 'as-is', without any express or implied warranty. In no event will the authors be held liable for any damages arising from the use of this software.

Permission is granted to anyone to use this software for any purpose, including commercial applications, and to alter it and redistribute it freely, subject to the following restrictions:

    1. The origin of this software must not be misrepresented; you must not claim that you wrote the original software. If you use this software in a product, an acknowledgment in the product documentation would be appreciated but is not required.

    2. Altered source versions must be plainly marked as such, and must not be misrepresented as being the original software.

    3. This notice may not be removed or altered from any source distribution.
*/
#include <windows.h>
#include <tchar.h>
#include <shlwapi.h>
#include "exdll.h"

extern "C" void __declspec(dllexport) FileReadBytes(HWND hwndParent, int string_size, 
                                      TCHAR *variables, stack_t **stacktop)
{
	EXDLL_INIT();
	
	TCHAR temp[16];
    popstring(temp);
	HANDLE hFile = HANDLE(_ttoi(temp));
    popstring(temp);
	DWORD bytes = _ttoi(temp);
    LPTSTR result = new TCHAR[bytes*2+1];
	LPBYTE resultBytes = LPBYTE(result);
	ReadFile(hFile, resultBytes, bytes, &bytes, NULL);
	result[bytes*2] = _T('\0');
	while (bytes--)
	{
		result[bytes*2+1] = _T("0123456789ABCDEF")[resultBytes[bytes]&15];
		result[bytes*2]   = _T("0123456789ABCDEF")[resultBytes[bytes]/16];
	}
	pushstring(result);
	delete[] result;
}

DWORD ConvertHexToBin(LPTSTR hex, BYTE** bin)
{
	*bin = new BYTE[lstrlen(hex)/2+1];
	LPBYTE scanBin = *bin;
	LPTSTR scanHex = hex;
	TCHAR ch;
	while ((ch = *scanHex++) != _T('\0'))
	{
		BYTE byte;
		if ((ch >= '0') && (ch <= '9'))
			byte = ch-'0';
		else if ((ch >= 'A') && (ch <= 'F'))
			byte = ch-'A'+10;
		else if ((ch >= 'a') && (ch <= 'f'))
			byte = ch-'a'+10;
		else
			continue;
		byte *= 16;
		ch = *scanHex++;
		if ((ch >= '0') && (ch <= '9'))
			byte |= ch-'0';
		else if ((ch >= 'A') && (ch <= 'F'))
			byte |= ch-'A'+10;
		else if ((ch >= 'a') && (ch <= 'f'))
			byte |= ch-'a'+10;
		else if (ch == _T('\0'))
			break;
		else
			continue;
		*scanBin++ = byte;
	}
	return scanBin-*bin;
}

extern "C" void __declspec(dllexport) HexToBin(HWND hwndParent, int string_size, 
                                      TCHAR *variables, stack_t **stacktop)
{
	EXDLL_INIT();
	
	LPTSTR hex = new TCHAR[string_size];
    popstring(hex);
	BYTE* bin;
	DWORD bytes = ConvertHexToBin(hex, &bin);
	bin[bytes] = '\0';
#ifdef _UNICODE
	MultiByteToWideChar(CP_ACP, 0, LPCSTR(bin), bytes+1, hex, string_size);
	pushstring(hex);
#else
	pushstring(LPCSTR(bin));
#endif
	delete[] hex;
	delete[] bin;
}

extern "C" void __declspec(dllexport) BinToHex(HWND hwndParent, int string_size, 
                                      TCHAR *variables, stack_t **stacktop)
{
	EXDLL_INIT();
	
	LPTSTR bin = new TCHAR[string_size];
    popstring(bin);
	DWORD count = lstrlen(bin);

    LPTSTR result = new TCHAR[count*2+1];
	result[count*2] = _T('\0');
	while (count--)
	{
		result[count*2+1] = _T("0123456789ABCDEF")[bin[count]&15];
		result[count*2]   = _T("0123456789ABCDEF")[(bin[count]/16)&15];
	}
	pushstring(result);
}

extern "C" void __declspec(dllexport) FileWriteBytes(HWND hwndParent, int string_size, 
                                      TCHAR *variables, stack_t **stacktop)
{
	EXDLL_INIT();
	
	TCHAR temp[16];
    popstring(temp);
	HANDLE hFile = HANDLE(_ttoi(temp));
	LPTSTR hex = new TCHAR[string_size];
    popstring(hex);
	BYTE* bin;
	DWORD bytes = ConvertHexToBin(hex, &bin);
	delete[] hex;
	WriteFile(hFile, bin, bytes, &bytes, NULL);
	delete [] bin;
	_itot(bytes, temp, 10);
	pushstring(temp);
}

extern "C" void __declspec(dllexport) FileFindBytes(HWND hwndParent, int string_size, 
                                      TCHAR *variables, stack_t **stacktop)
{
	EXDLL_INIT();
	
	TCHAR temp[16];
    popstring(temp);
	HANDLE hFile = HANDLE(_ttoi(temp));
	LPTSTR hex = new TCHAR[string_size];
    popstring(hex);
	BYTE* bin;
	DWORD binlen = ConvertHexToBin(hex, &bin);
	delete[] hex;
    popstring(temp);
	DWORD maxlen = _ttoi(temp);
	DWORD buflen = max(binlen*2, 16384);
	LPSTR buffer = new char[buflen];
	DWORD found = -1;
	DWORD read;
	do
	{
		read = 0;
		ReadFile(hFile, buffer, min(maxlen, buflen), &read, NULL);
		maxlen -= read;
		LPCSTR scan = buffer;
		while ((scan = (LPCSTR) memchr(scan, bin[0], buffer+read-scan)) != NULL)
		{
			DWORD left = min(binlen,DWORD(buffer+read-scan));
			if (memcmp(++scan, bin+1, left-1) == 0)
			{
				if (binlen <= left)
				{
					found = SetFilePointer(hFile, scan-1-buffer-read, NULL, FILE_CURRENT);
					break;
				}
				else // we stopped the comparison in the middle of our pattern
				{
					memmove(buffer, scan, left-1); // shift buffer to start (+1) of possible match
					scan = buffer;
					read = 0;
					ReadFile(hFile, buffer+left-1, min(maxlen, buflen-left+1), &read, NULL); // fill more of buffer from file
					maxlen -= read;
					read += left-1;
					if ((read+1 >= binlen) && (memcmp(buffer+left-1, bin+left, binlen-left) == 0))
					{
						found = SetFilePointer(hFile, 0-read-1, NULL, FILE_CURRENT);
						break;
					}
				}			
			}
		}
	} while ((found == -1) && (read != 0));
	delete[] buffer;
	delete[] bin;
	_itot(found, temp, 10);
	pushstring(temp);
}

extern "C" void __declspec(dllexport) FileTruncate(HWND hwndParent, int string_size, 
                                      TCHAR *variables, stack_t **stacktop)
{
	EXDLL_INIT();
	
	TCHAR temp[16];
    popstring(temp);
	HANDLE hFile = HANDLE(_ttoi(temp));
    SetEndOfFile(hFile);
}


extern "C" BOOL WINAPI DllMain(HINSTANCE hInstance, DWORD, LPVOID)
{
	return TRUE;
}
